import logging
import subprocess
from datetime import datetime
from os import chdir, getcwd, makedirs
from os.path import exists
from shutil import rmtree
from time import time
from typing import Optional, List, Dict

from rich.logging import RichHandler


def set_logger(logger_list: List[str], logger_level: Optional[Dict] = None):
    """
    This function will replace all handlers of each logger in ``logger_list`` with RichHandler.
    If there are some custom handlers in logger, they will be replaced too.

    :param logger_list: A list contains loggers.
    :type logger_list: list
    :param logger_level: You can specify the log level in ``logger_level``, with the name of logger is the key, and the level of logger is the value.
                         Default if None, with which all loggers' level will be set to ``logging.WARNING``.
    :type logger_level: list | None
    :return:
    :rtype:
    """
    formatter = logging.Formatter(
        "%(name)s :: %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    # use rich handler
    handler = RichHandler()
    handler.setFormatter(formatter)

    for logger_name in logger_list:
        if logger_name in logging.root.manager.loggerDict:
            _logger = logging.getLogger(logger_name)
            for _handler in _logger.handlers:
                _logger.removeHandler(_handler)
            _logger.addHandler(handler)

            if logger_level is not None and logger_name in logger_level:
                _logger.setLevel(logger_level[logger_name])
            else:
                _logger.setLevel(logging.WARNING)
                
                
# init wrfrun logger
logger = logging.getLogger("wrfrun")
set_logger(["wrfrun", ], {"wrfrun": logging.DEBUG})


def unify_logger_format():
    """
    This function is only supposed to be used internally.
    This function will replace all handlers of each logger with ``rich.logging.RichHandler``.
    Use this carefully.

    :return:
    :rtype:
    """
    set_logger(
        ["cdsapi", "cfgrib", "datapi"],
        {
            "cdsapi": logging.INFO,
            "cfgrib": logging.ERROR,
            "datapi": logging.INFO
        }
    )


def check_path(*args, force=False):
    """Check and create all the path in *args

    Returns:

    """
    for _path in args:
        if exists(_path) and force:
            rmtree(_path)
        if not exists(_path):
            makedirs(_path)


def logger_add_file_handler(log_path: str):
    """Add file handler to logger

    Args:
        log_path (str): Folder to place log file
    """
    # check log save path
    check_path(log_path)

    # add file handler
    file_handler = logging.FileHandler(
        f"{log_path}/{datetime.fromtimestamp(time()).strftime('%Y-%m-%d %H:%M:%S')}.log")
    file_handler.setFormatter(logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s :: %(message)s", datefmt="%m-%d %H:%M:%S"))
    logger.addHandler(file_handler)


def rectify_domain_size(point_num: int, nest_ratio: int) -> int:
    """Rectify domain size.

    Args:
        point_num (int): Point number of one side.
        nest_ratio (int): The nesting ratio relative to the domain’s parent.

    Returns:
        int: New size.
    """
    # calculate remainder
    point_num_mod = (point_num - 1) % nest_ratio

    if point_num_mod == 0:
        return point_num

    if point_num_mod < nest_ratio / 2:
        # # point_num_mod is closer to (nest_ratio - 1) than nest_ratio
        point_num -= point_num_mod
    else:
        # # point_num_mod is closer to nest_ratio than (nest_ratio - 1)
        point_num += (nest_ratio - point_num_mod)

    return point_num


def _calculate_domain_shape(step: float, resolution: int, grid_resolution=110, nest_ratio=1) -> int:
    """Calculate domain shape based on its area

    Args:
        step (float): Length of the side. Unit: degree.
        resolution (int): Resolution of domain. Unit: km.
        grid_resolution (int, optional): Resolution of grid. Unit: km. Defaults to 110.
        nest_ratio (int, optional): The nesting ratio relative to the domain’s parent.

    Returns:
        tuple[int, int]: (length of x, length of y)
    """
    # calculate res based on doamin area and resolution
    res = int(step * grid_resolution) // resolution

    # rectify

    return rectify_domain_size(res, nest_ratio=nest_ratio)


def calculate_domain_shape(lon_step: float, lat_step: float, resolution: int, grid_resolution=110, nest_ratio=1) -> tuple[int, int]:
    """Calculate domain shape based on its area

    Args:
        lon_step (float): Length in X direction. Unit: degree.
        lat_step (float): Length in Y direction. Unit: degree.
        resolution (int): Resolution of domain. Unit: km.
        grid_resolution (int, optional): Resolution of grid. Unit: km. Defaults to 110.
        nest_ratio (int, optional): The nesting ratio relative to the domain’s parent.

    Returns:
        tuple[int, int]: (length of x, length of y)
    """

    return (
        _calculate_domain_shape(
            lon_step, resolution, grid_resolution=grid_resolution, nest_ratio=nest_ratio),
        _calculate_domain_shape(
            lat_step, resolution, grid_resolution=grid_resolution, nest_ratio=nest_ratio)
    )


def _check_domain_shape(point_num: int, nest_ratio: int) -> bool:
    """Check domain shape.

    Args:
        point_num (int): Point number of one side.
        nest_ratio (int): The nesting ratio relative to the domain’s parent.

    Returns:
        bool: True if pass check else False.
    """
    if (point_num - 1) % nest_ratio != 0:
        return False
    else:
        return True


def check_domain_shape(x_point_num: int, y_point_num: int, nest_ratio: int) -> tuple[bool, bool]:
    """Check domain shape.

    Args:
        x_point_num (int): Point number of X side.
        y_point_num (int): Point number of Y side.
        nest_ratio (int): The nesting ratio relative to the domain’s parent.

    Returns:
        tuple[bool, bool]: Tuple of bool. True if pass check else False
    """
    return (
        _check_domain_shape(x_point_num, nest_ratio=nest_ratio),
        _check_domain_shape(y_point_num, nest_ratio=nest_ratio)
    )


def check_subprocess_status(status: subprocess.CompletedProcess):
    """Check subprocess return code and print log if their return code != 0

    Args:
        status (CompletedProcess): Status from subprocess.
    """
    if status.returncode != 0:
        # print command
        command = status.args
        logger.error(f"Failed to exec command: {command}")

        # print log
        logger.error(f"====== stdout ======")
        logger.error(status.stdout.decode())
        logger.error(f"====== ====== ======")
        logger.error(f"====== stderr ======")
        logger.error(status.stderr.decode())
        logger.error(f"====== ====== ======")

        # raise error
        raise RuntimeError


def call_subprocess(command: list[str], work_path: Optional[str] = None, print_output=False):
    """
    Execute the given command in system shell.

    :param command: A list contains the command and parameters to be executed.
    :type command: list
    :param work_path: The work path of the command.
                      If None, works in current directory.
    :type work_path: str | None
    :param print_output: If print standard output and error in the logger.
    :type print_output: bool
    :return:
    :rtype:
    """
    if work_path is not None:
        origin_path = getcwd()
        chdir(work_path)
    else:
        origin_path = None

    status = subprocess.run(' '.join(command), shell=True, capture_output=True)

    if origin_path is not None:
        chdir(origin_path)

    check_subprocess_status(status)

    if print_output:
        logger.info(status.stdout.decode())
        logger.warning(status.stderr.decode())


__all__ = ["logger", "check_path", "logger_add_file_handler", "calculate_domain_shape", "rectify_domain_size", "check_domain_shape",
           "check_subprocess_status", "call_subprocess", "set_logger", "unify_logger_format"]



from wrfrun.core import WRFRUNConfig
from wrfrun.utils import logger


def process_after_ndown():
    """
    After running ndown.exe, namelist settings are supposed to be changed,
    so WRF can simulate a higher resolution domain according to `WRF User's Guide <https://www2.mmm.ucar.edu/wrf/users/wrf_users_guide/build/html/running_wrf.html#wrf-nesting>`_.
    `wrfrun` provide this function to help you change these settings which have multiple values for each domain.
    The first value will be removed to ensure the value of higher resolution domain is the first value.

    :return:
    """
    namelist_data = WRFRUNConfig.get_namelist("wrf")

    for section in namelist_data:
        if section in ["bdy_control", "namelist_quilt"]:
            continue

        for key in namelist_data[section]:
            if key in ["grid_id", "parent_id", "i_parent_start", "j_parent_start", "parent_grid_ratio", "parent_time_step_ratio", "eta_levels"]:
                continue

            if isinstance(namelist_data[section][key], list):

                if len(namelist_data[section][key]) > 1:
                    namelist_data[section][key] = namelist_data[section][key][1:]

    namelist_data["domains"]["max_dom"] = 1

    time_ratio = namelist_data["domains"]["parent_time_step_ratio"][1]
    namelist_data["domains"]["time_step"] = namelist_data["domains"]["time_step"] // time_ratio

    WRFRUNConfig.update_namelist(namelist_data, "wrf")

    logger.info(f"Update namelist after running ndown.exe")


__all__ = ["process_after_ndown"]



"""
This file contains functions to interact with WRF workspace
"""

from os import listdir, makedirs, symlink
from os.path import exists

from .core import WRFRUNConfig
from .utils import check_path, logger


def prepare_workspace():
    """Initialize workspace

    """
    logger.info(f"Initialize workspace...")

    # extract WRF path
    wrf_config = WRFRUNConfig.get_model_config("wrf")
    wps_path = wrf_config["wps_path"]
    wrf_path = wrf_config["wrf_path"]
    wrfda_path = wrf_config["wrfda_path"]

    WRFRUN_TEMP_PATH = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRFRUN_TEMP_PATH)
    WORK_PATH = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRFRUN_WORKSPACE_PATH)
    REPLAY_WORK_PATH = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRFRUN_REPLAY_WORK_PATH)
    WPS_WORK_PATH = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WPS_WORK_PATH)
    WRF_WORK_PATH = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRF_WORK_PATH)
    WRFDA_WORK_PATH = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRFDA_WORK_PATH)
    output_path = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRFRUN_OUTPUT_PATH)

    # check folder
    check_path(WRFRUN_TEMP_PATH)
    check_path(REPLAY_WORK_PATH, force=True)
    check_path(output_path)

    if exists(wrfda_path):
        WRFRUNConfig.USE_WRFDA = True

    # create folder to run WPS, and WRF
    # check the path
    if not (exists(wps_path) and exists(wrf_path)):
        logger.error(f"Your WPS or WRF path is wrong")
        raise FileNotFoundError(f"Your WPS or WRF path is wrong")

    # remove old file
    if exists(WORK_PATH):
        logger.info(f"Remove old files...")
    check_path(WPS_WORK_PATH, f"{WPS_WORK_PATH}/outputs",
               WRF_WORK_PATH, WRFDA_WORK_PATH, force=True)
    logger.info(f"Link essential files...")

    # link {wps_path}/*
    # collect file except folder geogrid
    file_list = [
        x for x in listdir(wps_path) if x not in ["geogrid", "namelist.wps"]
    ]
    for file in file_list:
        symlink(f"{wps_path}/{file}", f"{WPS_WORK_PATH}/{file}")

    # create folder geogrid and link default GEOGRID file
    makedirs(f"{WPS_WORK_PATH}/geogrid")
    symlink(
        f"{wps_path}/geogrid/GEOGRID.TBL", f"{WPS_WORK_PATH}/geogrid/GEOGRID.TBL"
    )

    # # link {wrf_path}/run/*, except namelist.input
    file_list = [x for x in listdir(
        f"{wrf_path}/run") if not x.startswith("namelist")]
    for file in file_list:
        symlink(f"{wrf_path}/run/{file}", f"{WRF_WORK_PATH}/{file}")

    if WRFRUNConfig.USE_WRFDA:
        # # link {wrfda_path}/bin/*.exe
        file_list = ["da_wrfvar.exe", "da_update_bc.exe"]
        for file in file_list:
            symlink(f"{wrfda_path}/var/build/{file}", f"{WRFDA_WORK_PATH}/{file}")

        # # link {wrfda_path}/var/run/*
        file_list = listdir(f"{wrfda_path}/var/run")
        for file in file_list:
            symlink(f"{wrfda_path}/var/run/{file}", f"{WRFDA_WORK_PATH}/{file}")

        # # link {wrfda_path}/run/LANDUSE.TBL
        symlink(f"{wrfda_path}/run/LANDUSE.TBL", f"{WRFDA_WORK_PATH}/LANDUSE.TBL")


__all__ = ["prepare_workspace"]



import cfgrib as cf
import numpy as np
from pandas import to_datetime
from seafog import goos_sst_find_data, goos_sst_parser
from xarray import DataArray

from .utils import create_sst_grib
from wrfrun.core import WRFRUNConfig
from wrfrun.utils import logger


def merge_era5_goos_sst_grib(surface_grib_path: str, save_path: str, resolution="low"):
    """
    This function can read ERA5 skt data from the surface GRIB file, interpolate it to the same resolution as GOOS sst data, merge them, and save it to a new GRIB file.

    Args:
        surface_grib_path: ERA5 surface GRIB file path.
        save_path: The save path of the new GRIB file.
        resolution: Resolution of GOOS sst, valid value: ``["low", "high"]``.

    """
    dataset_list = cf.open_datasets(surface_grib_path)

    dataset = None
    for _ds in dataset_list:
        if "skt" in _ds:
            dataset = _ds
            break

    if dataset is None:
        logger.error(f"'skt' data not found in {surface_grib_path}")
        raise ValueError

    skt = dataset["skt"]

    longitude_start, longitude_end = skt["longitude"][0].data, skt["longitude"][-1].data
    latitude_start, latitude_end = skt["latitude"][-1].data, skt["latitude"][0].data

    data_time = skt["time"].to_numpy()
    data_time = to_datetime(data_time).strftime("%Y-%m-%d %H:%M")

    sst_data_path = WRFRUNConfig.get_model_config()["near_goos_data_folder"]

    # read the first time data, so we can interpolate skt data to the same resolution as sst data.
    _data = goos_sst_find_data(data_time[0], sst_data_path, resolution=resolution, show_progress=False)
    _data = goos_sst_parser(_data, resolution=resolution)
    _data = _data.loc[latitude_start:latitude_end, longitude_start:longitude_end]

    skt = skt.interp({"longitude": _data["longitude"], "latitude": _data["latitude"]})
    sst = np.zeros_like(skt)
    temp_data = np.zeros_like(skt[0])

    temp_data[:] = skt.to_numpy()[0]
    index = ~np.isnan(_data.to_numpy())
    temp_data[index] = _data.to_numpy()[index]
    sst[0] = temp_data

    # loop other time.
    for time_index, _data_time in enumerate(data_time[1:], start=1):
        _data = goos_sst_find_data(_data_time, sst_data_path, resolution=resolution, show_progress=False)
        _data = goos_sst_parser(_data, resolution=resolution)
        _data = _data.loc[latitude_start:latitude_end, longitude_start:longitude_end]

        temp_data[:] = skt.to_numpy()[time_index]
        index = ~np.isnan(_data.to_numpy())
        temp_data[index] = _data.to_numpy()[index]
        sst[time_index] = temp_data

    sst = DataArray(
        name="sst",
        data=sst,
        dims=["time", "latitude", "longitude"],
        coords={
            "time": skt["time"],
            "latitude": skt["latitude"],
            "longitude": skt["longitude"]
        }, attrs={
            "units": "K",
            "long_name": "Sea surface temperature"
        }
    )

    create_sst_grib(sst, save_path=save_path)


__all__ = ["merge_era5_goos_sst_grib"]



from os import listdir
from shutil import move

from ..core import WRFRUNConfig
from ..utils import check_path, logger


def clear_model_logs():
    """
    This function can automatically collect WRF log files and save them to ``output_path``.
    This function is used inside the wrfrun package.
    If you want to do something about the log files, check the corresponding code of interface functions in ``wrfrun.model.run``.

    :return:
    :rtype:
    """
    work_status = WRFRUNConfig.WRFRUN_WORK_STATUS
    work_path = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRF_WORK_PATH)

    log_files = [x for x in listdir(work_path) if x.startswith("rsl.")]

    if len(log_files) > 0:
        logger.warning(f"Found unprocessed log files of {work_status}")

        log_save_path = f"{WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRFRUN_OUTPUT_PATH)}/{work_status}/logs"
        check_path(log_save_path)

        for _file in log_files:
            move(f"{work_path}/{_file}", f"{log_save_path}/{_file}")

        logger.warning(f"Unprocessed log files of {work_status} has been saved to {log_save_path}, check it")


__all__ = ["clear_model_logs"]



from os import chdir, getcwd
from os.path import abspath, exists
from shutil import copyfile, move

from wrfrun.core import WRFRUNConfig
from wrfrun.res import EXT_NCL_PLOT_SCRIPT
from wrfrun.utils import call_subprocess, check_path, logger
from .wrf.namelist import prepare_wps_namelist
from .base import NamelistName


def plot_domain_area():
    """Generate namelist and plot domain area with WRF NCL script.

    """
    prepare_wps_namelist()

    # get save path
    save_path = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRFRUN_OUTPUT_PATH)

    # check
    check_path(save_path)

    # conver to absolute path
    save_path = abspath(save_path)
    save_path = f"{save_path}/wps_show_dom.png"

    # record original path
    origin_path = getcwd()

    # enter WPS WORK PATH
    chdir(WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WPS_WORK_PATH))

    # save namelist
    WRFRUNConfig.write_namelist(f"./{NamelistName.WPS}", "wps", overwrite=True)

    # copy plot script and plot
    copyfile(WRFRUNConfig.parse_resource_uri(EXT_NCL_PLOT_SCRIPT), f"./plotgrids.ncl")
    call_subprocess(["ncl", "./plotgrids.ncl"], print_output=True)

    # save image
    # we need to check image file, because sometimes ncl doesn't return error code
    if not exists("./wps_show_dom.png"):
        logger.error(f"Fail to plot domain with NCL. Check the log above")
        raise FileNotFoundError
    move("./wps_show_dom.png", save_path)

    logger.info(f"The image of domain area has been saved to {save_path}")

    # go back
    chdir(origin_path)


__all__ = ["plot_domain_area"]



import socket
import socketserver
import subprocess
from datetime import datetime
from time import time
from typing import Tuple, Optional

from .config import WRFRUNConfig
from ..utils import logger

WRFRUN_SERVER_INSTANCE = None
WRFRUN_SERVER_THREAD = None


def get_wrf_simulated_seconds(start_datetime: datetime, log_file_path: Optional[str] = None) -> int:
    """
    Get how many seconds wrf has integrated.

    :param start_datetime: WRF start datetime.
    :type start_datetime: datetime
    :param log_file_path: Absolute path of the log file to be parsed.
    :type log_file_path: str
    :return: Seconds.
    :rtype: int
    """
    # use linux cmd to get the latest line of wrf log files
    if log_file_path is None:
        log_file_path = WRFRUNConfig.parse_resource_uri(f"{WRFRUNConfig.WRF_WORK_PATH}/rsl.out.0000")
    res = subprocess.run(["tail", "-n", "1", log_file_path], capture_output=True)
    log_text = res.stdout.decode()

    if not (log_text.startswith("d01") or log_text.startswith("d02")):
        return -1

    time_string = log_text.split()[1]

    seconds = -1
    try:
        current_datetime = datetime.strptime(time_string, "%Y-%m-%d_%H:%M:%S")
        # remove timezone info so we can calculate.
        date_delta = current_datetime - start_datetime.replace(tzinfo=None)
        seconds = date_delta.days * 24 * 60 * 60 + date_delta.seconds

    except ValueError:
        seconds = -1

    finally:
        return seconds


class WRFRunServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """
    A socket server to report time usage.
    """

    def __init__(self, start_date: datetime, wrf_simulate_seconds: int, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # record the time the server starts
        self.start_timestamp = datetime.fromtimestamp(time())

        # record when the wrf start to integral
        self.start_date = start_date

        # record how many seconds the wrf will integral
        self.wrf_simulate_seconds = wrf_simulate_seconds

        # we need to parse the log file to track the simulation progress.
        self.wrf_log_path = WRFRUNConfig.parse_resource_uri(f"{WRFRUNConfig.WRF_WORK_PATH}/rsl.out.0000")
        logger.debug("WRFRun Server will try to track simulation progress with following log files:")
        logger.debug(f"WRF: {self.wrf_log_path}")

    def server_bind(self):
        """
        Bind address and port.
        """
        # reuse address and port to prevent the error `Address already in use`
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.server_address)

    def get_start_time(self) -> datetime:
        """
        Get the start time of the socket.
        As the socket server should start with NWP, it is also regarded as the start time of NWP.

        :return: Start datetime.
        :rtype: datetime
        """
        return self.start_timestamp

    def get_wrf_simulate_settings(self) -> Tuple[datetime, int]:
        """
        Get the start date of the case the NWP simulates and the total seconds of the simulation.

        :return: (start date, simulation seconds)
        :rtype: tuple
        """
        return self.start_date, self.wrf_simulate_seconds


class WRFRunServerHandler(socketserver.StreamRequestHandler):
    """
    Socket server handler.

    """
    def __init__(self, request, client_address, server: WRFRunServer) -> None:
        super().__init__(request, client_address, server)

        # get server
        self.server: WRFRunServer = server

    def calculate_time_usage(self) -> str:
        """Calculate time usage from server start (usually the time to run wrfrun)

        Returns:
            str: Time usage in `"%H:%M:%S"`
        """
        # get current timestamp
        current_timestamp = datetime.fromtimestamp(time())

        # get delta second
        seconds_diff = current_timestamp - self.server.get_start_time()
        seconds_diff = seconds_diff.seconds

        # calculate hours, minutes and seconds
        seconds = seconds_diff % 60
        minutes = (seconds_diff % 3600) // 60
        hours = seconds_diff // 3600

        time_usage = ":".join([
            str(hours).rjust(2, '0'),
            str(minutes).rjust(2, '0'),
            str(seconds).rjust(2, '0')
        ])

        return time_usage

    def calculate_progress(self) -> str:
        """
        Calculate the simulation progress.

        :return:
        :rtype:
        """
        start_date, simulate_seconds = self.server.get_wrf_simulate_settings()

        simulated_seconds = get_wrf_simulated_seconds(start_date, self.server.wrf_log_path)

        if simulated_seconds > 0:
            progress = simulated_seconds * 100 // simulate_seconds

        else:
            progress = 0

        status = WRFRUNConfig.WRFRUN_WORK_STATUS

        if status == "":
            status = "*"

        return f"{status}: {progress}%"

    def handle(self) -> None:
        """
        Request handler.
        """

        # check if we will to stop server
        msg = self.rfile.readline().decode().split('\n')[0]

        if msg == "stop":
            self.server.shutdown()
            self.wfile.write(f"Server stop\n".encode())
        
        elif msg == "debug":
            start_date, simulate_seconds = self.server.get_wrf_simulate_settings()
            start_date = start_date.strftime("%Y-%m-%d %H:%M")
            
            self.wfile.write(f"{start_date}\n{simulate_seconds}\n".encode())
            
        else:
            progress = self.calculate_progress()
            time_usage = self.calculate_time_usage()

            # send the message
            self.wfile.write(f"{progress}\n{time_usage}".encode())


def stop_server(socket_ip: str, socket_port: int):
    """Try to stop server.

    Args:
        socket_ip (str): Server IP.
        socket_port (int): Server Port.
    """
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        sock.connect((socket_ip, socket_port))

        # send msg to server
        sock.sendall("stop\n".encode())

        # receive the message
        msg = sock.recv(1024).decode().split('\n')[0]

        logger.info(f"WRFRunServer: {msg}")

    except (ConnectionRefusedError, ConnectionResetError):
        logger.warning("Fail to stop WRFRunServer, maybe it doesn't start at all.")


__all__ = ["WRFRunServer", "WRFRunServerHandler", "get_wrf_simulated_seconds", "stop_server"]



import subprocess
from enum import Enum
from json import dumps
from os import chdir, getcwd, listdir, makedirs, remove, symlink
from os.path import abspath, basename, dirname, exists, isdir
from shutil import copyfile, make_archive, move
from typing import Optional, TypedDict, Union

import numpy as np

from .config import WRFRUNConfig
from .error import CommandError, ConfigError, OutputFileError
from ..utils import check_path, logger


def check_subprocess_status(status: subprocess.CompletedProcess):
    """
    Check subprocess return code and print log if ``return_code != 0``.

    :param status: Status from subprocess.
    :type status: CompletedProcess
    """
    if status.returncode != 0:
        # print command
        command = status.args
        logger.error(f"Failed to exec command: {command}")

        # print log
        logger.error("====== stdout ======")
        logger.error(status.stdout.decode())
        logger.error("====== ====== ======")
        logger.error("====== stderr ======")
        logger.error(status.stderr.decode())
        logger.error("====== ====== ======")

        # raise error
        raise RuntimeError


def call_subprocess(command: list[str], work_path: Optional[str] = None, print_output=False):
    """
    Execute the given command in system shell.

    :param command: A list contains the command and parameters to be executed.
    :type command: list
    :param work_path: The work path of the command.
                      If None, works in current directory.
    :type work_path: str | None
    :param print_output: If print standard output and error in the logger.
    :type print_output: bool
    :return:
    :rtype:
    """
    if work_path is not None:
        origin_path = getcwd()
        chdir(work_path)
    else:
        origin_path = None

    status = subprocess.run(' '.join(command), shell=True, capture_output=True)

    if origin_path is not None:
        chdir(origin_path)

    check_subprocess_status(status)

    if print_output:
        logger.info(status.stdout.decode())
        logger.warning(status.stderr.decode())


def _json_default(obj):
    """
    Used for json.dumps.

    :param obj:
    :type obj:
    :return:
    :rtype:
    """
    if isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    else:
        raise TypeError(f"Object of type {type(obj)} is not JSON serializable.")


class InputFileType(Enum):
    """
    Input file type.
    ``WRFRUN_RES`` means the input file is from the NWP or wrfrun package.
    ``CUSTOM_RES`` means the input file is from the user, which may be a customized file.
    """
    WRFRUN_RES = 1
    CUSTOM_RES = 2


class FileConfigDict(TypedDict):
    """
    Dict class to give information to process files.
    """
    file_path: str
    save_path: str
    save_name: str
    is_data: bool
    is_output: bool


class ExecutableClassConfig(TypedDict):
    """
    Executable class initialization config template.
    """
    # only list essential config
    class_args: tuple
    class_kwargs: dict


class ExecutableConfig(TypedDict):
    """
    Executable config template.
    """
    name: str
    cmd: Union[str, list[str]]
    work_path: Optional[str]
    mpi_use: bool
    mpi_cmd: Optional[str]
    mpi_core_num: Optional[int]
    class_config: Optional[ExecutableClassConfig]
    input_file_config: Optional[list[FileConfigDict]]
    output_file_config: Optional[list[FileConfigDict]]
    custom_config: Optional[dict]


class _ExecutableConfigRecord:
    """
    Record executable configs and export them.
    """
    _instance = None
    _initialized = False

    def __init__(self, save_path: Optional[str] = None, include_data=False):
        """
        Record executable configs and export them.

        :param save_path: Save path of the exported config file.
        :type save_path: str
        :param include_data: If includes input data.
        :type include_data: bool
        """

        if self._initialized:
            return

        if save_path is None:
            WRFRUNConfig.IS_RECORDING = False
        else:
            WRFRUNConfig.IS_RECORDING = True

        self.save_path = save_path
        self.include_data = include_data

        self.work_path = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRFRUN_REPLAY_WORK_PATH)
        self.content_path = f"{self.work_path}/config_and_data"
        check_path(self.content_path)

        self._recorded_config = []
        self._name_count = {}

        self._initialized = True

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)

        return cls._instance

    def reinit(self, save_path: Optional[str] = None, include_data=False):
        """
        Reinitialize this instance.

        :return:
        :rtype:
        """
        self._initialized = False
        return _ExecutableConfigRecord(save_path, include_data)

    def record(self, exported_config: ExecutableConfig):
        """
        Record exported config for replay.

        :param exported_config: Executable config.
        :type exported_config: ExecutableConfig
        :return:
        :rtype:
        """
        if not self.include_data:
            self._recorded_config.append(exported_config)
            return

        check_path(self.content_path)

        # process exported config so we can also include data.
        # create directory to place data
        name = exported_config["name"]
        if name in self._name_count:
            self._name_count[name] += 1
            index = self._name_count[name]
        else:
            self._name_count[name] = 1
            index = 1

        data_save_uri = f"{WRFRUNConfig.WRFRUN_REPLAY_WORK_PATH}/{name}/{index}"
        data_save_path = f"{self.content_path}/{name}/{index}"
        makedirs(data_save_path)

        input_file_config = exported_config["input_file_config"]

        for _config_index, _config in enumerate(input_file_config):
            if not _config["is_data"]:
                continue

            if _config["is_output"]:
                continue

            file_path = _config["file_path"]
            file_path = WRFRUNConfig.parse_resource_uri(file_path)
            filename = basename(file_path)
            copyfile(file_path, f"{data_save_path}/{filename}")

            _config["file_path"] = f"{data_save_uri}/{filename}"
            input_file_config[_config_index] = _config

        exported_config["input_file_config"] = input_file_config
        self._recorded_config.append(exported_config)

    def clear_records(self):
        """
        Clean old configs.

        :return:
        :rtype:
        """
        self._recorded_config = []

    def export_replay_file(self):
        """
        Save replay file to the specific save path.

        :return:
        :rtype:
        """
        if len(self._recorded_config) == 0:
            logger.warning("No replay config has been recorded.")
            return

        logger.info("Exporting replay config... It may take a few minutes if you include data.")

        check_path(self.content_path)

        with open(f"{self.content_path}/config.json", "w") as f:
            f.write(dumps(self._recorded_config, indent=4, default=_json_default))

        if exists(self.save_path):
            if isdir(self.save_path):
                self.save_path = f"{self.save_path}/1.replay"
            else:
                if not self.save_path.endswith(".replay"):
                    self.save_path = f"{self.save_path}.replay"

                if exists(self.save_path):
                    logger.warning(f"Found existed replay file with the same name '{basename(self.save_path)}', overwrite it")
                    remove(self.save_path)

        if not exists(dirname(self.save_path)):
            makedirs(dirname(self.save_path))

        temp_file = f"{self.work_path}/config_and_data"
        make_archive(temp_file, "zip", self.content_path)
        move(f"{temp_file}.zip", self.save_path)

        logger.info(f"Replay config exported to {self.save_path}")


ExecConfigRecorder = _ExecutableConfigRecord()


class ExecutableBase:
    """
    Base class for all executables.
    """
    _instance = None

    def __init__(self, name: str, cmd: Union[str, list[str]], work_path: str, mpi_use=False, mpi_cmd: Optional[str] = None, mpi_core_num: Optional[int] = None):
        """
        Base class for all executables.

        :param name: Unique name to identify different executables.
        :type name: str
        :param cmd: Command to execute, can be a single string or a list contains the command and its parameters.
                    For example, ``"./geogrid.exe"``, ``["./link_grib.csh", "data/*", "."]``.
                    If you want to use mpi, then ``cmd`` must be a string.
        :type cmd: str
        :param work_path: Working directory path.
        :type work_path: str
        :param mpi_use: If you use mpi. You have to give ``mpi_cmd`` and ``mpi_core_num`` if you use mpi. Defaults to False.
        :type mpi_use: bool
        :param mpi_cmd: MPI command. For example, ``"mpirun"``. Defaults to None.
        :type mpi_cmd: str
        :param mpi_core_num: How many cores you use. Defaults to None.
        :type mpi_core_num: int
        """
        if mpi_use and isinstance(cmd, list):
            logger.error("If you want to use mpi, then `cmd` must be a single string.")
            raise CommandError("If you want to use mpi, then `cmd` must be a single string.")

        self.name = name
        self.cmd = cmd
        self.work_path = work_path
        self.mpi_use = mpi_use
        self.mpi_cmd = mpi_cmd
        self.mpi_core_num = mpi_core_num

        self.class_config: ExecutableClassConfig = {"class_args": (), "class_kwargs": {}}
        self.custom_config: dict = {}
        self.input_file_config: list[FileConfigDict] = []
        self.output_file_config: list[FileConfigDict] = []

        # directory to save outputs
        self._output_save_path = f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/{self.name}"
        self._log_save_path = f"{self._output_save_path}/logs"

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)

        return cls._instance

    def generate_custom_config(self):
        """
        Generate custom configs.

        :return:
        :rtype:
        """
        logger.debug(f"Method 'generate_custom_config' not implemented in '{self.name}'")

    def load_custom_config(self):
        """
        Load custom configs.

        :return:
        :rtype:
        """
        logger.debug(f"Method 'load_custom_config' not implemented in '{self.name}'")

    def export_config(self) -> ExecutableConfig:
        """
        Export config of this executable.

        :return: A dict contains configs.
        :rtype: ExecutableConfig
        """
        self.generate_custom_config()

        return {
            "name": self.name,
            "cmd": self.cmd,
            "work_path": self.work_path,
            "mpi_use": self.mpi_use,
            "mpi_cmd": self.mpi_cmd,
            "mpi_core_num": self.mpi_core_num,
            "class_config": self.class_config,
            "custom_config": self.custom_config,
            "input_file_config": self.input_file_config,
            "output_file_config": self.output_file_config
        }

    def load_config(self, config: ExecutableConfig):
        """
        Load config from a dict.

        :param config: Config dict. It must contain some essential keys. Check ``ExecutableConfig`` for details.
        :type config: ExecutableConfig
        :return:
        :rtype:
        """
        if "name" not in config:
            logger.error("A valid config is required. Please check ``ExecutableConfig``.")
            raise ValueError("A valid config is required. Please check ``ExecutableConfig``.")

        if self.name != config["name"]:
            logger.error(f"Config belongs to '{config['name']}', not {self.name}")
            raise ConfigError(f"Config belongs to '{config['name']}', not {self.name}")

        self.cmd = config["cmd"]
        self.work_path = config["work_path"]
        self.mpi_use = config["mpi_use"]
        self.mpi_cmd = config["mpi_cmd"]
        self.mpi_core_num = config["mpi_core_num"]
        self.class_config = config["class_config"]
        self.custom_config = config["custom_config"]
        self.input_file_config = config["input_file_config"]
        self.output_file_config = config["output_file_config"]

        self.load_custom_config()

    def replay(self):
        """
        This method will be called when replay the simulation.
        This method should take care every job that will be done when replaying the simulation.

        :return:
        :rtype:
        """
        logger.debug(f"Method 'replay' not implemented in '{self.name}', fall back to default action.")
        self()

    def add_input_files(self, input_files: Union[str, list[str], FileConfigDict, list[FileConfigDict]], is_data=True, is_output=True):
        """
        Add input files the extension will use.

        You can give a single file path or a list contains files' path.

        >>> self.add_input_files("data/custom_file")
        >>> self.add_input_files(["data/custom_file_1", "data/custom_file_2"])

        You can give more information with a ``FileConfigDict``, like the path and the name to store, and if it is data.

        >>> file_dict: FileConfigDict = {
        ...     "file_path": "data/custom_file.nc",
        ...     "save_path": f"{WRFRUNConfig.WPS_WORK_PATH}",
        ...     "save_name": "custom_file.nc",
        ...     "is_data": True,
        ...     "is_output": False
        ... }
        >>> self.add_input_files(file_dict)

        >>> file_dict_1: FileConfigDict = {
        ...     "file_path": "data/custom_file",
        ...     "save_path": f"{WRFRUNConfig.WPS_WORK_PATH}/geogrid",
        ...     "save_name": "GEOGRID.TBL",
        ...     "is_data": False,
        ...     "is_output": False
        ... }
        >>> file_dict_2: FileConfigDict = {
        ...     "file_path": "data/custom_file",
        ...     "save_path": f"{WRFRUNConfig.WPS_WORK_PATH}/outputs",
        ...     "save_name": "test_file",
        ...     "is_data": True,
        ...     "is_output": True
        ... }
        >>> self.add_input_files([file_dict_1, file_dict_2])

        :param input_files: Custom files' path.
        :type input_files: str | list | dict
        :param is_data: If its data. This parameter will be overwritten by the value in ``input_files``.
        :type is_data: bool
        :return:
        :rtype:
        """
        if isinstance(input_files, str):
            self.input_file_config.append(
                {
                    "file_path": input_files, "save_path": self.work_path, "save_name": basename(input_files),
                    "is_data": is_data, "is_output": is_output
                }
            )

        elif isinstance(input_files, list):
            for _file in input_files:
                if isinstance(_file, dict):
                    self.input_file_config.append(_file)  # type: ignore

                elif isinstance(_file, str):
                    self.input_file_config.append(
                        {
                            "file_path": _file, "save_path": self.work_path, "save_name": basename(_file),
                            "is_data": is_data, "is_output": is_output
                        }
                    )

                else:
                    logger.error(f"Input file config should be string or `FileConfigDict`, but got '{type(_file)}'")
                    raise TypeError(f"Input file config should be string or `FileConfigDict`, but got '{type(_file)}'")

        elif isinstance(input_files, dict):
            self.input_file_config.append(input_files)  # type: ignore

        else:
            logger.error(f"Input file config should be string or `FileConfigDict`, but got '{type(input_files)}'")
            raise TypeError(f"Input file config should be string or `FileConfigDict`, but got '{type(input_files)}'")

    def add_output_files(
        self, output_dir: Optional[str] = None, save_path: Optional[str] = None, startswith: Union[None, str, tuple[str, ...]] = None,
        endswith: Union[None, str, tuple[str, ...]] = None, outputs: Union[None, str, list[str]] = None, no_file_error=True
    ):
        """
        Add save file rules.

        :param output_dir: Output dir paths.
        :type output_dir: str
        :param save_path: Save path.
        :type save_path: str
        :param startswith: Prefix string or prefix list of output files.
        :type startswith: str | list
        :param endswith: Postfix string or Postfix list of output files.
        :type endswith: str | list
        :param outputs: Files name list. All files in the list will be saved.
        :type outputs: str | list
        :param no_file_error: If True, an error will be raised with the ``error_message``. Defaults to True.
        :type no_file_error: bool
        :return:
        :rtype:
        """
        if WRFRUNConfig.FAKE_SIMULATION_MODE:
            return

        if output_dir is None:
            output_dir = self.work_path

        if save_path is None:
            save_path = f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/{self.name}"

        file_list = listdir(WRFRUNConfig.parse_resource_uri(output_dir))
        save_file_list = []

        if startswith is not None:
            _list = []
            for _file in file_list:
                if _file.startswith(startswith):
                    _list.append(_file)
            save_file_list += _list

            logger.debug(f"Collect files match `startswith`: {_list}")

        if endswith is not None:
            _list = []
            for _file in file_list:
                if _file.endswith(endswith):
                    _list.append(_file)
            save_file_list += _list

            logger.debug(f"Collect files match `endswith`: {_list}")

        if outputs is not None:
            if isinstance(outputs, str) and outputs in file_list:
                save_file_list.append(outputs)
            else:
                outputs = [x for x in outputs if x in file_list]
                save_file_list += outputs

        if len(save_file_list) < 1:
            if no_file_error:
                logger.error(f"Can't find any files match the giving rules: startswith='{startswith}', endswith='{endswith}', outputs='{outputs}'")
                raise OutputFileError(f"Can't find any files match the giving rules: startswith='{startswith}', endswith='{endswith}', outputs='{outputs}'")

            else:
                logger.warning(f"Can't find any files match the giving rules: startswith='{startswith}', endswith='{endswith}', outputs='{outputs}'. Skip it.")
                return

        save_file_list = list(set(save_file_list))
        logger.debug(f"Files to be processed: {save_file_list}")

        for _file in save_file_list:
            self.output_file_config.append({"file_path": f"{output_dir}/{_file}", "save_path": save_path, "save_name": _file, "is_data": True, "is_output": True})

    def before_exec(self):
        """
        Prepare input files.

        :return:
        :rtype:
        """
        if WRFRUNConfig.FAKE_SIMULATION_MODE:
            logger.info(f"We are in fake simulation mode, skip preparing input files for '{self.name}'")
            return

        for input_file in self.input_file_config:
            file_path = input_file["file_path"]
            save_path = input_file["save_path"]
            save_name = input_file["save_name"]

            file_path = WRFRUNConfig.parse_resource_uri(file_path)
            save_path = WRFRUNConfig.parse_resource_uri(save_path)

            file_path = abspath(file_path)
            save_path = abspath(save_path)

            if not exists(file_path):
                logger.error(f"File not found: '{file_path}'")
                raise FileNotFoundError(f"File not found: '{file_path}'")

            if not exists(save_path):
                makedirs(save_path)

            target_path = f"{save_path}/{save_name}"
            if exists(target_path):
                logger.debug(f"Target file {save_name} exists, overwrite it.")
                remove(target_path)

            symlink(file_path, target_path)

    def after_exec(self):
        """
        Save outputs and logs.
        
        :return: 
        :rtype: 
        """
        if WRFRUNConfig.FAKE_SIMULATION_MODE:
            logger.info(f"We are in fake simulation mode, skip saving outputs for '{self.name}'")
            return

        for output_file in self.output_file_config:
            file_path = output_file["file_path"]
            save_path = output_file["save_path"]
            save_name = output_file["save_name"]

            file_path = WRFRUNConfig.parse_resource_uri(file_path)
            save_path = WRFRUNConfig.parse_resource_uri(save_path)

            file_path = abspath(file_path)
            save_path = abspath(save_path)

            if not exists(file_path):
                logger.error(f"File not found: '{file_path}'")
                raise FileNotFoundError(f"File not found: '{file_path}'")

            if not exists(save_path):
                makedirs(save_path)

            target_path = f"{save_path}/{save_name}"
            if exists(target_path):
                logger.warning(f"Found existed file, which means you already may have output files in '{save_path}'. If you are saving logs, ignore this warning.")

            move(file_path, target_path)

    def exec(self):
        """
        Execute the given command.

        :return:
        :rtype:
        """
        work_path = WRFRUNConfig.parse_resource_uri(self.work_path)

        if not self.mpi_use or None in [self.mpi_cmd, self.mpi_core_num]:
            if isinstance(self.cmd, str):
                self.cmd = [self.cmd, ]

            logger.info(f"Running `{' '.join(self.cmd)}` ...")
            _cmd = self.cmd

        else:
            logger.info(f"Running `{self.mpi_cmd} --oversubscribe -np {self.mpi_core_num} {self.cmd}` ...")
            _cmd = [self.mpi_cmd, "--oversubscribe", "-np", str(self.mpi_core_num), self.cmd]

        if WRFRUNConfig.FAKE_SIMULATION_MODE:
            logger.info(f"We are in fake simulation mode, skip calling numerical model for '{self.name}'")
            return

        call_subprocess(_cmd, work_path=work_path)

    def __call__(self):
        """
        Execute the given command.

        :return:
        :rtype:
        """
        self.before_exec()
        self.exec()
        self.after_exec()

        if not WRFRUNConfig.IS_IN_REPLAY and WRFRUNConfig.IS_RECORDING:
            ExecConfigRecorder.record(self.export_config())


__all__ = ["ExecutableBase", "FileConfigDict", "InputFileType", "ExecutableConfig", "ExecutableClassConfig", "ExecConfigRecorder"]



from datetime import datetime, timedelta
from os.path import exists
from typing import Union

from wrfrun.core import WRFRUNConfig
from wrfrun.res import NAMELIST_DFI, NAMELIST_WPS, NAMELIST_WRF, NAMELIST_WRFDA
from wrfrun.utils import logger
from .scheme import *


def _check_start_end_date(max_dom: int, start_date: Union[datetime, list[datetime]], end_date: Union[datetime, list[datetime]]) -> tuple[list[datetime], list[datetime]]:
    """
    Format start date and end date.

    :param max_dom: Domain number.
    :type max_dom: int
    :param start_date: Date list parsed from the config file.
    :type start_date: datetime | list
    :param end_date: Date list parsed from the config file.
    :type end_date: datetime | list
    :return: Formated date list.
    :rtype: list
    """
    if isinstance(start_date, datetime):
        start_date = [start_date for _ in range(max_dom)]
    elif isinstance(start_date, list):
        if len(start_date) != max_dom:
            logger.error(f"You have {max_dom} domains, but you only give {len(start_date)} dates for `start_date`.")
            raise ValueError(f"You have {max_dom} domains, but you only give {len(start_date)} dates for `start_date`.")

    if isinstance(end_date, datetime):
        end_date = [end_date for _ in range(max_dom)]
    elif isinstance(end_date, list):
        if len(end_date) != max_dom:
            logger.error(f"You have {max_dom} domains, but you only give {len(end_date)} dates for `start_date`.")
            raise ValueError(f"You have {max_dom} domains, but you only give {len(end_date)} dates for `start_date`.")

    return start_date, end_date


def prepare_wps_namelist():
    """
    This function read WPS template namelist and update its value based on the config file and user custom namelist.

    """
    # prepare namelist
    WRFRUNConfig.read_namelist(WRFRUNConfig.parse_resource_uri(NAMELIST_WPS), "wps")
    wrf_config = WRFRUNConfig.get_model_config("wrf")

    # get domain number
    max_dom = wrf_config["domain"]["domain_num"]

    # get start_date and end_date
    start_date = wrf_config["time"]["start_date"]
    end_date = wrf_config["time"]["end_date"]

    start_date, end_date = _check_start_end_date(max_dom, start_date, end_date)
    start_date = [x.strftime("%Y-%m-%d_%H:%M:%S") for x in start_date]
    end_date = [x.strftime("%Y-%m-%d_%H:%M:%S") for x in end_date]

    # get input data time interval
    interval_seconds = wrf_config["time"]["input_data_interval"]

    # generate update settings based on the config file
    update_value = {
        "share": {
            "max_dom": max_dom,
            "start_date": start_date,
            "end_date": end_date,
            "interval_seconds": interval_seconds
        },
        "geogrid": {
            "parent_grid_ratio": wrf_config["domain"]["parent_grid_ratio"],
            "i_parent_start": wrf_config["domain"]["i_parent_start"],
            "j_parent_start": wrf_config["domain"]["j_parent_start"],
            "e_we": wrf_config["domain"]["e_we"],
            "e_sn": wrf_config["domain"]["e_sn"],
            "dx": wrf_config["domain"]["dx"],
            "dy": wrf_config["domain"]["dy"],
            "ref_lat": wrf_config["domain"]["ref_lat"],
            "ref_lon": wrf_config["domain"]["ref_lon"],
            "stand_lon": wrf_config["domain"]["stand_lon"],
            "geog_data_path": wrf_config["geog_data_path"]
        }
    }

    # # use loop to process config of map_proj
    for key in wrf_config["domain"]["map_proj"]:
        if key == "name":
            update_value["geogrid"]["map_proj"] = wrf_config["domain"]["map_proj"][key]
        else:
            update_value["geogrid"][key] = wrf_config["domain"]["map_proj"][key]

    # # update namelist
    WRFRUNConfig.update_namelist(update_value, "wps")

    # # update settings from custom namelist
    if wrf_config["user_wps_namelist"] != "" and exists(wrf_config["user_wps_namelist"]):
        WRFRUNConfig.update_namelist(wrf_config["user_wps_namelist"], "wps")


def prepare_wrf_namelist():
    """
    This function read WRF template namelist and update its value based on the config file and user custom namelist.

    """
    # read template namelist
    WRFRUNConfig.read_namelist(WRFRUNConfig.parse_resource_uri(NAMELIST_WRF), "wrf")

    # wrf config from config
    wrf_config = WRFRUNConfig.get_model_config("wrf")

    # get debug level
    debug_level = WRFRUNConfig["model"]["debug_level"]

    # get domain number, start_date and end_date
    max_dom = wrf_config["domain"]["domain_num"]
    start_date = wrf_config["time"]["start_date"]
    end_date = wrf_config["time"]["end_date"]
    
    start_date, end_date = _check_start_end_date(max_dom, start_date, end_date)

    # get the time interval of input data and output data
    input_data_interval = wrf_config["time"]["input_data_interval"]
    output_data_interval = wrf_config["time"]["output_data_interval"]

    # get restart settings
    restart = wrf_config["restart_mode"]
    restart_interval = wrf_config["time"]["restart_interval"]
    if restart_interval < 0:
        restart_interval = output_data_interval

    # get the time step of integral
    time_step = wrf_config["time"]["time_step"]

    # calculate run hours
    run_hours = end_date[0] - start_date[0]
    run_hours = run_hours.days * 24 + run_hours.seconds // 3600

    # calculate dx and dy for each domain
    dx = wrf_config["domain"]["dx"]
    dy = wrf_config["domain"]["dy"]
    parent_grid_ratio = wrf_config["domain"]["parent_grid_ratio"]
    dx = [dx // ratio for ratio in parent_grid_ratio]
    dy = [dy // ratio for ratio in parent_grid_ratio]

    # prepare update values
    update_values = {
        "time_control": {
            # make sure run days, minutes, and seconds are 0
            "run_days": 0,
            "run_minutes": 0,
            "run_seconds": 0,
            "run_hours": run_hours,
            "start_year": [_date.year for _date in start_date],
            "start_month": [_date.month for _date in start_date],
            "start_day": [_date.day for _date in start_date],
            "start_hour": [_date.hour for _date in start_date],
            "start_minute": [_date.minute for _date in start_date],
            "start_second": [_date.second for _date in start_date],
            "end_year": [_date.year for _date in end_date],
            "end_month": [_date.month for _date in end_date],
            "end_day": [_date.day for _date in end_date],
            "end_hour": [_date.hour for _date in end_date],
            "end_minute": [_date.minute for _date in end_date],
            "end_second": [_date.second for _date in end_date],
            "interval_seconds": input_data_interval,
            "history_interval": [output_data_interval for _ in range(max_dom)],
            "auxinput4_interval": [input_data_interval // 60 for _ in range(max_dom)],
            "restart": restart,
            "restart_interval": restart_interval,
            "debug_level": debug_level,
        },
        "domains": {
            "max_dom": max_dom,
            "time_step": time_step,
            "parent_grid_ratio": parent_grid_ratio,
            "i_parent_start": wrf_config["domain"]["i_parent_start"],
            "j_parent_start": wrf_config["domain"]["j_parent_start"],
            "e_we": wrf_config["domain"]["e_we"],
            "e_sn": wrf_config["domain"]["e_sn"],
            "dx": dx,
            "dy": dy,

        },
        "physics": {}
    }

    # and we need to check the physics scheme option
    long_wave_scheme = {
        "ra_lw_physics": [SchemeLongWave.get_scheme_id(wrf_config["scheme"]["long_wave_scheme"]["name"]) for _ in range(max_dom)]
    }
    # # and other related options
    long_wave_scheme.update(wrf_config["scheme"]["long_wave_scheme"]["option"])
    # update
    update_values["physics"].update(long_wave_scheme)

    short_wave_scheme = {
        "ra_sw_physics": [SchemeShortWave.get_scheme_id(wrf_config["scheme"]["short_wave_scheme"]["name"]) for _ in range(max_dom)]
    }
    # # and other related options
    short_wave_scheme.update(
        wrf_config["scheme"]["short_wave_scheme"]["option"])
    # update
    update_values["physics"].update(short_wave_scheme)

    cumulus_scheme = {
        "cu_physics": [SchemeCumulus.get_scheme_id(wrf_config["scheme"]["cumulus_scheme"]["name"]) for _ in range(max_dom)]
    }
    # # and other related options
    cumulus_scheme.update(wrf_config["scheme"]["cumulus_scheme"]["option"])
    # update
    update_values["physics"].update(cumulus_scheme)

    pbl_scheme = {
        "bl_pbl_physics": [SchemePBL.get_scheme_id(wrf_config["scheme"]["pbl_scheme"]["name"]) for _ in range(max_dom)]
    }
    # # and other related options
    pbl_scheme.update(wrf_config["scheme"]["pbl_scheme"]["option"])
    # update
    update_values["physics"].update(pbl_scheme)

    land_surface_scheme = {
        "sf_surface_physics": [SchemeLandSurfaceModel.get_scheme_id(wrf_config["scheme"]["land_surface_scheme"]["name"]) for _ in range(max_dom)]
    }
    # # and other related options
    land_surface_scheme.update(
        wrf_config["scheme"]["land_surface_scheme"]["option"])
    # update
    update_values["physics"].update(land_surface_scheme)

    surface_layer_scheme = {
        "sf_sfclay_physics": [SchemeSurfaceLayer.get_scheme_id(wrf_config["scheme"]["surface_layer_scheme"]["name"]) for _ in range(max_dom)]
    }
    # # and other related options
    surface_layer_scheme.update(
        wrf_config["scheme"]["surface_layer_scheme"]["option"])
    # update
    update_values["physics"].update(surface_layer_scheme)

    # update namelist
    WRFRUNConfig.update_namelist(update_values, "wrf")

    # read user real namelist and update value
    user_namelist_data = wrf_config["user_wrf_namelist"]
    if user_namelist_data != "" and exists(user_namelist_data):
        WRFRUNConfig.update_namelist(user_namelist_data, "wrf")


def prepare_dfi_namelist():
    """Generate namelist data for DFI running

    """
    # Read template namelist
    WRFRUNConfig.read_namelist(WRFRUNConfig.parse_resource_uri(NAMELIST_DFI), "dfi")

    wrf_config = WRFRUNConfig.get_model_config("wrf")

    # Read start date and end date
    start_date = wrf_config["time"]["start_date"]
    start_date = start_date[0] if isinstance(start_date, list) else start_date
    input_data_interval = wrf_config["time"]["input_data_interval"]
    time_step = wrf_config["time"]["time_step"]
    # calculate dfi date because:
    # dfi start date is 1 hour earlier than start_date
    # dfi end date is 30 minutes later than start_date
    dfi_start_date = start_date - timedelta(hours=1)
    dfi_end_date = start_date + timedelta(minutes=30)

    # calculate dx and dy for each domain
    dx = wrf_config["domain"]["dx"]
    dy = wrf_config["domain"]["dy"]
    parent_grid_ratio = wrf_config["domain"]["parent_grid_ratio"]
    dx = [dx // ratio for ratio in parent_grid_ratio]
    dy = [dy // ratio for ratio in parent_grid_ratio]

    # Construct update value
    update_value = {
        "time_control": {
            # make sure run days, hours, minutes, seconds is 0
            "run_days": 0,
            "run_hours": 0,
            "run_minutes": 0,
            "run_seconds": 0,
            # start date and end date are same
            "start_year": [start_date.year],
            "start_month": [start_date.month],
            "start_day": [start_date.day],
            "start_hour": [start_date.hour],
            "start_minute": [start_date.minute],
            "start_second": [start_date.second],
            "end_year": [start_date.year],
            "end_month": [start_date.month],
            "end_day": [start_date.day],
            "end_hour": [start_date.hour],
            "end_minute": [start_date.minute],
            "end_second": [start_date.second],
            "interval_seconds": input_data_interval,
            "auxinput4_interval_s": [input_data_interval, ]
        },
        "domains": {
            # make sure max_dom = 1
            "time_step": time_step,
            "max_dom": 1,
            "time_step_dfi": time_step if time_step < 90 else 90,
            "parent_grid_ratio": parent_grid_ratio,
            "i_parent_start": wrf_config["domain"]["i_parent_start"],
            "j_parent_start": wrf_config["domain"]["j_parent_start"],
            "e_we": wrf_config["domain"]["e_we"],
            "e_sn": wrf_config["domain"]["e_sn"],
            "dx": dx,
            "dy": dy,
        },
        "dfi_control": {
            # set dfi date
            "dfi_bckstop_year": dfi_start_date.year,
            "dfi_bckstop_month": dfi_start_date.month,
            "dfi_bckstop_day": dfi_start_date.day,
            "dfi_bckstop_hour": dfi_start_date.hour,
            "dfi_bckstop_minute": dfi_start_date.minute,
            "dfi_bckstop_second": dfi_start_date.second,
            "dfi_fwdstop_year": dfi_end_date.year,
            "dfi_fwdstop_month": dfi_end_date.month,
            "dfi_fwdstop_day": dfi_end_date.day,
            "dfi_fwdstop_hour": dfi_end_date.hour,
            "dfi_fwdstop_minute": dfi_end_date.minute,
            "dfi_fwdstop_second": dfi_end_date.second,
        }
    }

    # update namelist data
    WRFRUNConfig.update_namelist(update_value, "dfi")

    # read user wrf namelist and update value
    user_namelist_data = wrf_config["user_wrf_namelist"]
    if user_namelist_data != "" and exists(user_namelist_data):
        WRFRUNConfig.update_namelist(user_namelist_data, "dfi")


def prepare_wrfda_namelist():
    """Generate namelist for da_wrfvar.exe

    """
    # read template namelist
    WRFRUNConfig.read_namelist(WRFRUNConfig.parse_resource_uri(NAMELIST_WRFDA), "wrfda")

    wrf_config = WRFRUNConfig.get_model_config("wrf")

    # get wrf start date
    start_date = wrf_config["time"]["start_date"]
    start_date = start_date[0] if isinstance(start_date, list) else start_date

    # generate update value
    update_value = {
        "wrfvar18": {
            "analysis_date": start_date.strftime("%Y-%m-%d_%H:%M:%S.0000")
        },
        "wrfvar21": {
            # one hour before wrf start date
            "time_window_min": (start_date - timedelta(hours=1)).strftime("%Y-%m-%d_%H:%M:%S.0000")
        },
        "wrfvar22": {
            # one hour after wrf start date
            "time_window_max": (start_date + timedelta(hours=1)).strftime("%Y-%m-%d_%H:%M:%S.0000")
        }
    }

    # update namelist
    WRFRUNConfig.update_namelist(update_value, "wrfda")

    # read user wrfda namelist and update value
    user_namelist_data = wrf_config["user_wrfda_namelist"]
    if user_namelist_data != "" and exists(user_namelist_data):
        WRFRUNConfig.update_namelist(user_namelist_data, "wrfda")


__all__ = ["prepare_wrf_namelist", "prepare_wps_namelist", "prepare_wrfda_namelist", "prepare_dfi_namelist"]



from os.path import basename
from typing import Optional, Union

from wrfrun import WRFRUNConfig
from .core import DFI, GeoGrid, MetGrid, Real, UnGrib, WRF, NDown


def geogrid(geogrid_tbl_file: Union[str, None] = None):
    """
    Interface to execute geogrid.exe.

    :param geogrid_tbl_file: Custom GEOGRID.TBL file path. Defaults to None.
    """
    GeoGrid(geogrid_tbl_file, WRFRUNConfig.get_core_num())()


def ungrib(vtable_file: Union[str, None] = None, input_data_path: Optional[str] = None, prefix="FILE"):
    """
    Interface to execute ungrib.exe.

    :param vtable_file: Vtable file used to run ungrib. Defaults to None.
    :type vtable_file: str
    :param input_data_path: Directory path of the input data. If None, ``wrfrun`` will read its value from the config file.
    :type input_data_path: str
    :param prefix: Prefix of ungrib output.
    :type prefix: str
    """
    prefix = basename(prefix)
    WRFRUNConfig.set_ungrib_out_prefix(prefix)

    UnGrib(vtable_file, input_data_path)()


def metgrid(geogrid_data_path: Optional[str] = None, ungrib_data_path: Optional[str] = None, fg_names: Union[str, list[str]] = "FILE"):
    """
    Interface to execute metgrid.exe.

    :param geogrid_data_path: Directory path of outputs from geogrid.exe. If None, tries to use the output path specified by config file.
    :type geogrid_data_path: str
    :param ungrib_data_path: Directory path of outputs from ungrib.exe. If None, tries to use the output path specified by config file.
    :type ungrib_data_path: str
    :param fg_names: Set ``fg_name`` of metgrid, a single prefix string or a string list.
    :type fg_names: str | list
    """
    if isinstance(fg_names, str):
        fg_names = [fg_names, ]
    fg_names = [basename(x) for x in fg_names]
    WRFRUNConfig.set_metgrid_fg_names(fg_names)

    MetGrid(geogrid_data_path, ungrib_data_path, WRFRUNConfig.get_core_num())()


def real(metgrid_data_path: Union[str, None] = None):
    """
    Interface to execute real.exe.

    :param metgrid_data_path: The path store output from metgrid.exe. If it is None, the default output path will be used.
    """
    Real(metgrid_data_path, WRFRUNConfig.get_core_num())()


def wrf(input_file_dir_path: Union[str, None] = None, restart_file_dir_path: Optional[str] = None, save_restarts=False):
    """
    Interface to execute wrf.exe.

    :param input_file_dir_path: The path store input data which will be feed into wrf.exe. Defaults to None.
    :param restart_file_dir_path: The path store WRF restart files. This parameter will be ignored if ``restart=False`` in your config.
    :param save_restarts: Also save restart files to the output directory.
    """
    WRF(input_file_dir_path, restart_file_dir_path, save_restarts, WRFRUNConfig.get_core_num())()


def dfi(input_file_dir_path: Optional[str] = None, update_real_output=True):
    """
    Execute "wrf.exe" to run DFI.

    :param input_file_dir_path: Path of the directory that stores input data for "wrf.exe".
    :type input_file_dir_path: str
    :param update_real_output: If update the corresponding file in real.exe output directory.
    :type update_real_output: bool
    """
    DFI(input_file_dir_path, update_real_output, WRFRUNConfig.get_core_num())()


def ndown(wrfout_file_path: str, real_output_dir_path: Optional[str] = None, update_namelist=True):
    """
    Execute "ndown.exe".

    :param wrfout_file_path: wrfout file path.
    :type wrfout_file_path: str
    :param real_output_dir_path: Path of the directory that contains output of "real.exe".
    :type real_output_dir_path: str
    :param update_namelist: If update wrf's namelist for the final integral.
    :type update_namelist: bool
    :return:
    :rtype:
    """
    NDown(wrfout_file_path, real_output_dir_path, update_namelist, WRFRUNConfig.get_core_num())()


__all__ = ["geogrid", "ungrib", "metgrid", "real", "wrf", "dfi", "ndown"]



from typing import Tuple

from cartopy import crs
from cartopy.mpl.geoaxes import GeoAxes
from matplotlib.collections import QuadMesh
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure
from netCDF4 import Dataset  # type: ignore
from wrf import getvar, latlon_coords, get_cartopy, to_np
from xarray import DataArray

from wrfrun.utils import logger

# Here defines the colormap for landuse and soiltype
LANDUSE_CMAP = ListedColormap([
    [0, .4, 0],         # 1 Evergreen Needleleaf Forest
    [0, .4, .2],        # 2 Evergreen Broadleaf Forest
    [.2, .8, .2],       # 3 Deciduous Needleleaf Forest
    [.2, .8, .4],       # 4 Deciduous Broadleaf Forest
    [.2, .6, .2],       # 5 Mixed Forests
    [.3, .7, 0],        # 6 Closed Shrublands
    [.82, .41, .12],    # 7 Open Shurblands
    [.74, .71, .41],    # 8 Woody Savannas
    [1, .84, .0],       # 9 Savannas
    [0, 1, 0],          # 10 Grasslands
    [0, 1, 1],          # 11 Permanant Wetlands
    [1, 1, 0],          # 12 Croplands
    [1, 0, 0],          # 13 Urban and Built-up
    [.7, .9, .3],       # 14 Cropland/Natual Vegation Mosaic
    [1, 1, 1],          # 15 Snow and Ice
    [.914, .914, .7],   # 16 Barren or Sparsely Vegetated
    [.5, .7, 1],        # 17 Water (like oceans)
    [.86, .08, .23],    # 18 Wooded Tundra
    [.97, .5, .31],     # 19 Mixed Tundra
    [.91, .59, .48],    # 20 Barren Tundra
    [0, 0, .88],        # 21 Lake
])

LANDUSE_LABELS = [
    'Evergreen Needleleaf Forest',
    'Evergreen Broadleaf Forest',
    'Deciduous Needleleaf Forest',
    'Deciduous Broadleaf Forest',
    'Mixed Forests',
    'Closed Shrublands',
    'Open Shrublands',
    'Woody Savannas',
    'Savannas',
    'Grasslands',
    'Permanent Wetlands',
    'Croplands',
    'Urban and Built-Up',
    'Cropland/Natural Vegetation Mosaic',
    'Snow and Ice',
    'Barren or Sparsely Vegetated',
    'Water',
    'Wooded Tundra',
    'Mixed Tundra',
    'Barren Tundra',
    'Lake',
]


# Here defines the colormap and labels for soil type
SOILTYPE_CMAP = ListedColormap([
    [0, .4, 0],         # 1 Sand
    [0, .4, .2],        # 2 Loamy Sand
    [.2, .8, .2],       # 3 Sandy Loam
    [.2, .8, .4],       # 4 Silt Loam
    [.2, .6, .2],       # 5 Silt
    [.3, .7, 0],        # 6 Loam
    [.82, .41, .12],    # 7 Sandy Clay Loam
    [.74, .71, .41],    # 8 Silty Clay Loam
    [1, .84, .0],       # 9 Clay Loam
    [0, 1, 0],          # 10 Sandy Clay
    [0, 1, 1],          # 11 Silty Clay
    [1, 1, 0],          # 12 Clay
    [1, 0, 0],          # 13 Organic Material
    [.7, .9, .3],       # 14 Water
    [1, 1, 1],          # 15 Bedrock
    [.914, .914, .7]    # 16 Other
])

SOILTYPE_LABELS = [
    "Sand",
    "Loamy Sand",
    "Sandy Loam",
    "Silt Loam",
    "Silt",
    "Loam",
    "Sandy Clay Loam",
    "Silty Clay Loam",
    "Clay Loam",
    "Sandy Clay",
    "Silty Clay",
    "Clay",
    "Organic Material",
    "Water",
    "Bedrock",
    "Other"
]


def get_cmap_ticks(name: str) -> tuple:
    """Get corresponding colormap, labels and ticks.

    Args:
        name (str): Field name.

    Returns:
        tuple: (colormap, labels, ticks)
    """
    # name map
    cmap_ticks_map = {
        "LU_INDEX": {
            "cmap": LANDUSE_CMAP,
            "labels": LANDUSE_LABELS,
        },
        "SOILCTOP": {
            "cmap": SOILTYPE_CMAP,
            "labels": SOILTYPE_LABELS,
        },
        "SOILCBOT": {
            "cmap": SOILTYPE_CMAP,
            "labels": SOILTYPE_LABELS,
        },
    }

    # take out colormap and labels
    if name not in cmap_ticks_map:
        logger.error(f"Can't found colormap and labels for {name}")
        raise ValueError

    cmap = cmap_ticks_map[name]["cmap"]
    labels = cmap_ticks_map[name]["labels"]

    # generate ticks
    ticks = [x + 1.5 for x in range(len(labels))]

    return cmap, labels, ticks


def draw_geogrid(data_path: str, field: str, fig: Figure, nrow: int, ncol: int, index: int) -> Tuple[GeoAxes, QuadMesh]:
    """Draw specific data in geo_em*

    Args:
        data_path (str): data path.
        field (str): Which field you want to draw.
        fig (Figure): Matplotlib figure object.
        nrow (int): Row number of axes.
        ncol (int): Column number of axes.
        index (int): Index of axes.
    """
    # take out data
    data: DataArray = getvar(Dataset(data_path), field)     # type: ignore

    # soiltype data is different from landuse data
    if field in ["SOILCTOP", "SOILCBOT"]:
        data = data.argmax(dim="soil_cat", keep_attrs=True)     # type: ignore
        # get data attrs
        data_attrs = data.attrs
        data = data + 1
        data.attrs = data_attrs

    # get latitude and longitude
    lats, lons = latlon_coords(data)

    # get projection
    proj = get_cartopy(data)

    # add axes
    ax = fig.add_subplot(nrow, ncol, index, projection=proj)

    # get colormap, labels and ticks
    cmap, labels, ticks = get_cmap_ticks(field)

    # convert data to numpy
    data, lats, lons = to_np(data), to_np(lats), to_np(lons)    # type: ignore

    # plot
    im = ax.pcolormesh(
        lons, lats, data, vmin=1, vmax=len(labels) + 1, cmap=cmap, transform=crs.PlateCarree()
    )

    return ax, im   # type: ignore


__all__ = ["draw_geogrid", "get_cmap_ticks"]
